{%- extends "basic/layout.html" %}
{%- block extrahead %}
  {{ super() }}
  {% if theme_touch_icon %}
  <link rel="apple-touch-icon" href="{{ pathto('_static/' ~ theme_touch_icon, 1) }}" />
  {% endif %}
  <link media="only screen and (max-device-width: 480px)" href="{{
    pathto('_static/small_flask.css', 1) }}" type= "text/css" rel="stylesheet" />
{% endblock %}
{%- block relbar2 %}
  {% if theme_github_fork %}
    <a href="http://github.com/{{ theme_github_fork }}"><img style="position: fixed; top: 0; right: 0; border: 0;"
    src="http://s3.amazonaws.com/github/ribbons/forkme_right_darkblue_121621.png" alt="Fork me on GitHub" /></a>
  {% endif %}
{% endblock %}
{% block header %}
  {{ super() }}
  {% if pagename == 'index' %}
  <div>
  {% endif %}
{% endblock %}
{%- block footer %}
  <div class="footer">
    &copy; Copyright {{ copyright }}. numbakit-ode {{ version }}.
    Created using <a href="http://sphinx.pocoo.org/">Sphinx</a>.
  </div>
  {% if pagename == 'index' %}
  </div>
  {% endif %}
{%- endblock %}
